<?php

namespace Teknisa\SM\Service;

use Doctrine\ORM\EntityManager;
use Teknisa\SM\Factory\Sm00022000CalculoFormacaoPrecoFactory as CalcFormPrecFactory;
use Teknisa\Libs\Util\Environment;
use Teknisa\Libs\Util\NewCode;

class Sm00022000CalculoFormacaoPrecoService {
    
    protected $connection;
    protected $CalcFormPrecFactory;
    protected $Environment;
    protected $calculados;
    protected $tabela;
    protected $nritem;
    
    public function __construct(EntityManager $entityManager, CalcFormPrecFactory $CalcFormPrecFactory, Environment $Environment,  NewCode $newCode) {
        $this->connection             = $entityManager->getConnection();
        $this->CalcFormPrecFactory = $CalcFormPrecFactory;
        $this->Environment  = $Environment;
        $this->newCode                = $newCode;
    }
    
    /**
     * guarda o valor da tabela e do numero do item
    **/
    public function setVrTabelaItem($tabela, $nritem){
        $this->tabela = $tabela;
        $this->nritem = $nritem;
    }
    
    /**
     * Função que verifica qual tabela esta sendo manipulada e retorna os parametros certos
    **/
    public function verificaTabela($nrFormPrecoServico, $nrCalcFormPrecServ){
        $params = null;
        if($this->tabela == "FCT_CALCFORMPRECSERV_ITPRO"){
            $params = array(
                array(
                    "NRFORMPRECOSERVICO" => $nrFormPrecoServico,
                    "NRORG"              => $this->Environment->getNrOrg(),
                    "NRITEMORCAMENTO"    => $this->nritem
                ),
                array(
                    "NRCALCFORMPRECSERV" => $nrCalcFormPrecServ,
                    "NRORG"              => $this->Environment->getNrOrg(),
                    "NRITEMORCAMENTO"    => $this->nritem
                ),
            );
        } else {
            if($this->tabela == "FCT_CALCFORMPRECSERV_ITCONT"){
                $params = array(
                    array(
                        "NRFORMPRECOSERVICO" => $nrFormPrecoServico,
                        "NRORG"              => $this->Environment->getNrOrg(),
                        "NRITEMOCONTRATO"    => $this->nritem
                    ),
                    array(
                        "NRCALCFORMPRECSERV" => $nrCalcFormPrecServ,
                        "NRORG"              => $this->Environment->getNrOrg(),
                        "NRITEMOCONTRATO"    => $this->nritem
                    )
                );
            } else {
                $params = array(
                    array(
                        "NRFORMPRECOSERVICO" => $nrFormPrecoServico,
                        "NRORG"              => $this->Environment->getNrOrg()
                    ),
                    array(
                        "NRCALCFORMPRECSERV" => $nrCalcFormPrecServ,
                        "NRORG"              => $this->Environment->getNrOrg()
                    )
                );
            }
        }
        
        return $params;
    }
    
   /**
     * A função "recebeOrdenacao" recebe a ordenação máxima para cálculo.
     */
     
     public function recebeOrdenacao($nrFormPrecoServico){
        $ordenacao = array();
        
        $params = $this->verificaTabela($nrFormPrecoServico, null)[0];
        if($this->tabela == "FCT_CALCFORMPRECSERV_ITPRO")
            $ordenacao = $this->CalcFormPrecFactory->recebeOrdenacaoAux($params);
        else
            if($this->tabela == "FCT_CALCFORMPRECSERV_ITCONT")
                $ordenacao = $this->CalcFormPrecFactory->recebeOrdenacaoAuxCont($params);
            else 
                $ordenacao = $this->CalcFormPrecFactory->recebeOrdenacao($params);

        return isset($ordenacao) ? $ordenacao[0] : null;
    }
    
    /**
     *  A função "validaIdentificador"  Realiza a validação  do campo "identificador". 
     *  Não podem haver dois identificadores iguais em uma mesma composição.
     */
     
     public function validaIdentificador($nrFormPrecoServico, $IdItemCalculo, $nrCalcFormPrecServ){
        $params = $this->verificaTabela($nrFormPrecoServico, null)[0];
        $demaisParams = array(
            "IDITEMCALCULO"      => $IdItemCalculo,
            "NRCALCFORMPRECSERV" => $nrCalcFormPrecServ
        );
        
        $params = array_merge($params, $demaisParams);
        $valIdentificador = array();
        
        if($this->tabela == "FCT_CALCFORMPRECSERV_ITPRO")
            $valIdentificador = $this->CalcFormPrecFactory->validaIdentificadorAux($params);
        else
            if($this->tabela == "FCT_CALCFORMPRECSERV_ITCONT")
                $valIdentificador = $this->CalcFormPrecFactory->validaIdentificadorAuxCont($params);
            else 
                $valIdentificador = $this->CalcFormPrecFactory->validaIdentificador($params);
        
        return isset($valIdentificador) ? $valIdentificador[0] : null;
    }
    
    /**
     *  A função "buscaIdentValores" busca os identificadores dos elementos e seus respectivos valores
     */
     public function buscaIdentValores($nrFormPrecoServico, $IdItemCalculo){
        $demaisParams = array(
            "IDITEMCALCULO"      => $IdItemCalculo,
        );
        $params = array_merge($this->verificaTabela($nrFormPrecoServico, null)[0], $demaisParams);
        $IdentValores = array();
        if($this->tabela == "FCT_CALCFORMPRECSERV_ITPRO")
            $IdentValores = $this->CalcFormPrecFactory->buscaIdentValoresAux($params);
        else 
            if($this->tabela == "FCT_CALCFORMPRECSERV_ITCONT")
                $IdentValores = $this->CalcFormPrecFactory->buscaIdentValoresCont($params);
            else 
                $IdentValores = $this->CalcFormPrecFactory->buscaIdentValores($params);

        return count($IdentValores) > 0 ? $IdentValores[0] : null;
    }
    
    
    
    /**
     *  A função "buscaIdentValoresAnte" busca os identificadores dos elementos e seus respectivos valores
     */
     
     public function buscaIdentValoresAnte($nrFormPrecoServico){
    
        $params = array(
            "NRFORMPRECOSERVICO" => $nrFormPrecoServico,
            "NRORG"              => $this->Environment->getNrOrg()
        );
        
        $IdentValores = $this->CalcFormPrecFactory->buscaIdentValoresAnte($params);
        
        return $IdentValores;
    }
    
    /**
     *  A função "calcSomatorioFilhos" calcula o somatorio dos elementos filhos 
     */
     
     public function calcSomatorioFilhos($nrCalcFormPrecServ, $tableName){
        $calcSumFilhos = array();
        if($tableName == "FCT_CALCFORMPRECSERV_ITPRO")
            $calcSumFilhos = $this->CalcFormPrecFactory->calcSomatorioFilhos_ItPro($this->verificaTabela(null, $nrCalcFormPrecServ)[1]);
        else 
            if($tableName == "FCT_CALCFORMPRECSERV_ITCONT")
                $calcSumFilhos = $this->CalcFormPrecFactory->calcSomatorioFilhos_ItCont($this->verificaTabela(null, $nrCalcFormPrecServ)[1]);
            else 
                $calcSumFilhos = $this->CalcFormPrecFactory->calcSomatorioFilhos($this->verificaTabela(null, $nrCalcFormPrecServ)[1]);
        $vrSoma = $calcSumFilhos[0]["VRSOMA"];
        $params = array(
            "NRCALCFORMPRECSERV" => $nrCalcFormPrecServ,
            "VRSOMA"             => $vrSoma,
            "NRORG"              => $this->Environment->getNrOrg()
        );
       $this->CalcFormPrecFactory->updSomatorioFilhos($params);
    }
    
    /**
     *  A função "deletarPaieFilhos"  deleta os pais e os filhos.
     */
     
     public function deletarPaieFilhos($nrCalcFormPrecServ){
    
        $params = array(
            "NRCALCFORMPRECSERV" => $nrCalcFormPrecServ,
            "NRORG"              => $this->Environment->getNrOrg()
        );
        
        $this->CalcFormPrecFactory->deletarPaieFilhos($params);
        
    }
    
     /**
     *  A função "recalculoElemento"  inicia os recalculos dos elementos.
     */
     
    public function recalculoElemento($nrCalcFormPrecServ) {
        $this->calculados = array();
        $this->recalculoElementoAux($nrCalcFormPrecServ, null);
    }

    /**
     *  A função "recalcularArvore"  Inicia o recalculo de toda a árvore de elementos.
     */
    public function recalcularArvore($nrFormPrecoServico) {
        $this->calculados = array();
        $params = $this->verificaTabela($nrFormPrecoServico, null)[0];
        $elementos = array();
        
        if($this->tabela == "FCT_CALCFORMPRECSERV_ITPRO")
            $elementos = $this->CalcFormPrecFactory->getElementosArvoreCalcFormProp($params);
        else 
            if($this->tabela == "FCT_CALCFORMPRECSERV_ITCONT")
                $elementos = $this->CalcFormPrecFactory->getElementosArvoreCalcFormCont($params);
            else 
                $elementos = $this->CalcFormPrecFactory->getElementosArvoreCalcFormPrev($params);
           
        foreach ($elementos as $elemento) {
            $this->recalculoElementoAux($elemento["NRCALCFORMPRECSERV"], $this->tabela);
        }
    }
    
    /**
     *  A função "recalculoElemento" realiza o recalculo dos elementos.
     */
    
     public function recalculoElementoAux($nrCalcFormPrecServ) {
        
        /* 1 - Busca os Elementos para Recálculo  (Somente os do tipo "Somatório dos Filhos" e "Expreessão")*/
        $params = $this->verificaTabela(null, $nrCalcFormPrecServ)[1];
        $arrayRecalculo = array();
        
        if($this->tabela == "FCT_CALCFORMPRECSERV_ITPRO")
            $arrayRecalculo = $this->CalcFormPrecFactory->getElementoCalcFormPrecProp($params);
        else 
            if($this->tabela == "FCT_CALCFORMPRECSERV_ITCONT")
                $arrayRecalculo = $this->CalcFormPrecFactory->getElementoCalcFormPrecCont($params);
            else 
                $arrayRecalculo = $this->CalcFormPrecFactory->getElementoCalcFormPrec($params);
        
        if(count($arrayRecalculo) <= 0)
            throw new \Exception("Não há elementos filhos ou expressões para a realização do cálculo do elemento.");
        
        $recalculo = $arrayRecalculo[0];
        
        /* 2 - Realiza o Recalculo dos Elementos */        
        $nrFormPrecoServico = $recalculo["NRFORMPRECOSERVICO"];
        $nrCalcFormPrecServ = $recalculo["NRCALCFORMPRECSERV"];
        $quantidade         = $recalculo["QTCALCULO"];
        $usaMedida          = $recalculo["IDUSAMEDIDA"];
        $qtMedida           = $recalculo["QTMEDIDACALC"];
        $vrPercentual       = $recalculo["VRPERCCALCULO"];
        $idItemCalculo      = $recalculo["IDITEMCALCULO"];
        
        if (array_key_exists($idItemCalculo, $this->calculados)) {
            return $this->calculados[$idItemCalculo];
        } else {
            /* - Somatório - */        
            if ($recalculo["IDTIPOCALCULO"] == "S") {
                return $this->somatorioRecalculoCalc($nrCalcFormPrecServ, $idItemCalculo, $quantidade, $usaMedida, $qtMedida);
            } else {
            /* - Expressão - */
               $dsExpessaoCalc = $recalculo["DSEXPRESSAOCALC"];
               return $this->recalcularExpressao($nrFormPrecoServico, $nrCalcFormPrecServ, $idItemCalculo, $dsExpessaoCalc, $quantidade, $usaMedida, $qtMedida, $vrPercentual);
                
            } 
        }
        
    }
    /**
     *  A função "somatorioRecalculo" realiza o somatorio dos filhos para recalculo dos elementos.
     */
    public function somatorioRecalculoCalc($nrCalcFormPrecServ, $idItemCalculo, $quantidade, $usaMedida, $qtMedida){
        if (array_key_exists($idItemCalculo, $this->calculados)) {
            return $this->calculados[$idItemCalculo];
        } else {
            $params              = $this->verificaTabela(null, $nrCalcFormPrecServ)[1];
            $elementosFilhos     = array();
            $valorUnitario       = 0;
            $somatorioPercentual = 0;
            
            if($this->tabela == "FCT_CALCFORMPRECSERV_ITPRO")
                $elementosFilhos = $this->CalcFormPrecFactory->getElementosFilhosCalcFormPrecProp($params);
            else
                if($this->tabela == "FCT_CALCFORMPRECSERV_ITCONT")
                    $elementosFilhos = $this->CalcFormPrecFactory->getElementosFilhosCalcFormPrecCont($params);
                else 
                    $elementosFilhos = $this->CalcFormPrecFactory->getElementosFilhosCalcFormPrec($params);
            
            foreach ($elementosFilhos as $elemento) {
                if (array_key_exists($elemento["IDITEMCALCULO"], $this->calculados)) {
                    $valorUnitario       += $this->calculados[$elemento["IDITEMCALCULO"]]["VRTOTAL"];
                    $somatorioPercentual += $this->calculados[$elemento["IDITEMCALCULO"]]["VRPERCCALCULO"];
                } else {
                    if ($elemento["IDTIPOCALCULO"] == "V") {
                        $valorUnitario += (float)$elemento["VRTOTAL"];
                        $somatorioPercentual += (float)$elemento["VRPERCCALCULO"];
                    } else if ($elemento["IDTIPOCALCULO"] == "S") {
                        $resultado = $this->somatorioRecalculoCalc($elemento["NRCALCFORMPRECSERV"], $elemento["IDITEMCALCULO"], $elemento["QTCALCULO"], $elemento["IDUSAMEDIDA"], $elemento["QTMEDIDACALC"]);
                        
                        $valorUnitario       += $resultado["VRTOTAL"];
                        $somatorioPercentual += $resultado["VRPERCCALCULO"];
                    } else {
                        $resultado = $this->recalcularExpressao($elemento["NRFORMPRECOSERVICO"], $elemento["NRCALCFORMPRECSERV"], $elemento["IDITEMCALCULO"], $elemento["DSEXPRESSAOCALC"], $elemento["QTCALCULO"], $elemento["IDUSAMEDIDA"], $elemento["QTMEDIDACALC"], $elemento["VRPERCCALCULO"]);
                        
                        $valorUnitario       += $resultado["VRTOTAL"];
                        $somatorioPercentual += $resultado["VRPERCCALCULO"];
                    }
                }
            }
            
            $valorTotal = $valorUnitario * $quantidade;
            if ($usaMedida == "S") 
                $valorTotal *= $qtMedida;
            
            $this->calculados[$idItemCalculo]["VRTOTAL"] = $valorTotal;
            $this->calculados[$idItemCalculo]["VRPERCCALCULO"] = $somatorioPercentual;
            /* Salva o recálculo do Elemento */
            $this->salvarRecalculo($nrCalcFormPrecServ, $valorUnitario, $quantidade, $qtMedida, $valorTotal, $somatorioPercentual);  
            return $this->calculados[$idItemCalculo];
        }
    }
    
     
    /**
     *  A função "recalcularExpressao" realiza o recalculo dos valores para elementos do tipo "Expressão".
     */
    public function recalcularExpressao($nrFormPrecoServico, $nrCalcFormPrecServ, $idItemCalculo, $expressao, $quantidade, $usaMedida, $qtMedida, $vrPercentual) {
        
        if (array_key_exists($idItemCalculo, $this->calculados)) {
            return $this->calculados[$idItemCalculo];
        } else {
            $expressao = str_replace(array("[", "]"), "@@", $expressao);
            $ids = explode("@@", $expressao);
            
            for ($i = 1; $i < count($ids); $i = $i + 2) {
                $elementoExp = $this->buscaIdentValores($nrFormPrecoServico, $ids[$i]);
                if (!empty($elementoExp)) {
                    $nrCalcFormPrecServElementoExp = $elementoExp["NRCALCFORMPRECSERV"];
                    $vrIdTipoCalculoElementoExp    = $elementoExp["IDTIPOCALCULO"];
                    $idItemCalculoElementoExp      = $elementoExp["IDITEMCALCULO"];
                    $expressaoElementoExp          = $elementoExp["DSEXPRESSAOCALC"];
                    $percentualElementoExp         = $elementoExp["VRPERCCALCULO"];
                    $quantidadeElementoExp         = $elementoExp["QTCALCULO"];
                    $usaMedidaElementoExp          = $elementoExp["IDUSAMEDIDA"];
                    $qtMedidaElementoExp           = $elementoExp["QTMEDIDACALC"];
                    
                    if ($vrIdTipoCalculoElementoExp == "E") {
                        $resultado = $this->recalculoElementoAux($nrCalcFormPrecServElementoExp);
                        $valorTotalElementoExp = $resultado["VRTOTAL"];
                        
                    } else if ($vrIdTipoCalculoElementoExp == "S") {
                        $resultado = $this->somatorioRecalculoCalc($nrCalcFormPrecServElementoExp, $idItemCalculoElementoExp, $quantidadeElementoExp, $usaMedidaElementoExp, $qtMedidaElementoExp);
                        $valorTotalElementoExp = $resultado["VRTOTAL"];
                    } else {
                        $valorTotalElementoExp = $elementoExp["VRTOTAL"];
                    }

                    $expressao = str_replace("@@".$ids[$i]."@@", $valorTotalElementoExp, $expressao);
                }
            }
            
            /* Realiza o cálculo do elemento */    
            $vrUnitario = eval('return ' . $expressao . ';');
            if (!empty($vrPercentual)) {
                $vrUnitario = $vrUnitario * $vrPercentual / 100;
            }
            
            $vrTotal = $vrUnitario * $quantidade;
            if ($usaMedida == 'S') {
                $vrTotal *= $qtMedida; 
            }
            
            $this->calculados[$idItemCalculo]["VRTOTAL"] = $vrTotal;
            $this->calculados[$idItemCalculo]["VRPERCCALCULO"] = $vrPercentual;
            
            /* Salva o recálculo do Elemento */
            $this->salvarRecalculo($nrCalcFormPrecServ, $vrUnitario, $quantidade, $qtMedida, $vrTotal, $vrPercentual);
            return $this->calculados[$idItemCalculo];
        }
    }
    
    
    /**
     *  A função "salvarRecalculo" salva o recalculo dos elementos.
     */
     
     public function salvarRecalculo($nrCalcFormPrecServ, $vrUnitario, $qtCalculo, $qtmedida, $vrTotal, $vrPercentual){
    
        $demaisParams = array(
            "VRUNITARIO"         => $vrUnitario,
            "QTCALCULO"          => $qtCalculo,
            "QTMEDIDACALC"       => $qtmedida,
            "VRTOTAL"            => $vrTotal,
            "VRPERCCALCULO"      => $vrPercentual,
        );
        
        $params = array_merge($this->verificaTabela(null, $nrCalcFormPrecServ)[1], $demaisParams);
        if($this->tabela == "FCT_CALCFORMPRECSERV_ITPRO")
            $this->CalcFormPrecFactory->salvarRecalculoProp($params); 
        else
            if($this->tabela == "FCT_CALCFORMPRECSERV_ITCONT")
                $this->CalcFormPrecFactory->salvarRecalculoCont($params); 
            else 
                $this->CalcFormPrecFactory->salvarRecalculo($params); 
    }
    
    /**
     * A função "cadastramentoAutomatico realiza o cadastramento automático 
     * dos itens do cálculo da formação de preço
     */
    public function cadastramentoAutomatico($formacao, $idElemento, $idElementoPai, $idItemCalculo){
        $nrorg    = $this->Environment->getNrOrg();
        $operador = $this->Environment->getCdOperador();
        
        $this->connection->beginTransaction();
        $itAux = array();
        
        try {
            /* 1 - Insere todos os itens na tabela FCT_CALCFORMPRECSERV*/
            $count = count($idElemento["IDELEMENTO"]);
            for ($i = 0; $i < $count; $i++) {
            
                 $novoCodigo = $this->newCode->getCodeByTableName("FCT_CALCFORMPRECSERV", $nrorg, 1, 10, true);
                 $ident      = $idItemCalculo[$i];
                 $referencia = '['.$ident.'] - '.$idElemento["NMSERVORCA"][$i];
                 $pai        = 'CAD: '.$idElemento["IDELEMENTOPAI"][$i];
                 $params = array(
                    "NRCALCFORMPRECSERV"    => $novoCodigo,
                    "NRFORMPRECOSERVICO"    => $formacao,
                    "IDORDENACAO"           => $idElemento['IDORDENAELEMENTO'][$i],
                    "IDITEMCALCULO"         => $ident,
                    "IDELEMENTO"            => $idElemento["IDELEMENTO"][$i],
                    "NRCALCFORMPRECSERVPAI" => NULL,
                    "DSREFECALCULO"         => $referencia,
                    "CDOPERULTATU"          => $pai,
                    "NRORG"                 => $nrorg,
                    "CDOPERADOR"            => $operador, 
                    "DSEXPRESSAOCALC"       => $idElemento["DSEXPRESSAOCALC"][$i],
                    "IDTIPOCALCULO"         => $idElemento["IDTIPOCALCULO"][$i],
                    "VRPERCCALCULO"         => $idElemento["VRPERCCALCULO"][$i] 
                );
                
                $this->CalcFormPrecFactory->insereItensCadAuto($params);
                $itAux[$i]["PAI"] = $idElemento["IDELEMENTO"][$i];
                $itAux[$i]["COD"] = $novoCodigo;
            }
            
            /* 2 - Realiza a Hierarquização dos itens. */
            
            $idElementoPai = array_unique($idElementoPai);
            $idElementoPai = array_filter($idElementoPai);
            
            foreach($idElementoPai as $elementoPai){
                /* 2.1 - busca o código do Item pai */
                foreach($itAux as $aux){
                    if ($aux["PAI"] ==  $elementoPai){
                        
                        $codPai = $aux["COD"];
                    }
                }
                
                /*  2.2 - Substitui no Item pai */
                $expressao = 'CAD: '.$elementoPai;
                
                $params = array(
                    "NRCALCFORMPRECSERVPAI" => $codPai,
                    "NRFORMPRECOSERVICO"    => $formacao,
                    "CDOPERULTATU"          => $expressao,
                    "NRORG"                 => $nrorg,
                );
                
                $this->CalcFormPrecFactory->atualizaItensCadAuto($params);        
            }
            
            /*2.3 - Limpa o campo de expressão utilizado no cadastramento automatico */
            $params = array(
                "NRFORMPRECOSERVICO" => $formacao,
                "NRORG"              => $nrorg,
            );
            
            $this->CalcFormPrecFactory->limpaExpressao($params);  
            $this->connection->commit();
            
        }
        catch (\Exception $e) {
            $this->connection->rollback();
            throw $e;
        }
    }
    
}

?>